#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
/* For testing */
#include <time.h>


#include "combat.h"
#include "game.h"
#include "log.h"
#include "map.h"
#include "mem.h"
#include "rand.h"

#include "maphand.h"
#include "settlers.h"
#include "unithand.h"
#include "unittools.h"

#include "path_finding.h"

/* For speed set to powers of 2 */
#define MAX_XSIZE map.xsize
#define MAX_YSIZE map.ysize

/*======================================================================*/
/*===================== Hidden structures ==============================*/
/*======================================================================*/
typedef short xy_t;

struct pf_node {
  u_char cost;                    /* total cost up to now */
  u_char extra_cost;              /* total extra cost up to now */
  enum direction8 dir_to_here;
  enum pf_node_status {
    NS_UNINIT = 0,                /* memory is calloced, hence zero 
                                   * means uninitialised */
    NS_NEW,                       /* the optimal route isn't found yet */
    NS_PROCESSED                  /* the optimal route is found */
  } status;
  enum known_type node_known_type;
  enum tile_behaviour behaviour; 
  bool is_my_zoc;
  bool is_allied_tile;
  int extra;                      /* extra cost of the tile */
};
typedef struct pf_node *pf_node_t;

#define PQDATUM xy_t
#define PQDATUM_NULL -1
#define INIT_ITEMS 100

struct pqueue {
  int size;        /* number of occupied cells */
  int avail;       /* total number of cells */
  int step;        /* additional memory allocation step */
  PQDATUM *cells;     /* contain data */
  int *priorities; /* backup priorities (in case data is changed 
                    * while in the queue */
};


/* 
 * Generic move cost function:
 * (x,y) is the target tile, dir is the direction to the target tile.
 */
typedef int (COSTFN)(pf_map_t the_map, int x, int y, int dir);

/*
 * Anything else goes here?
 */
struct pf_map {
  short x, y, xy;                      /* The current position */
  COSTFN *get_cost;                    /* The move cost function */
  struct pf_parameter *params;
  struct pqueue *queue;
  struct pf_node *lattice;
};

/* ------ The queue implementation --------- */

/**********************************************************************
 * pqinit: initialize the queue.
 *
 * n is the numer of queue items for which memory should be 
 * preallocated, that is, the initial size of the item array the queue 
 * uses. If you insert more than n items to the queue, another n items 
 * will be allocated automatically.
 **********************************************************************/
struct pqueue *pqcreate(int n)
{
  struct pqueue *q;

  q = fc_malloc(sizeof(struct pqueue));
  q->cells = fc_malloc(sizeof(PQDATUM) * n);
  q->priorities = fc_malloc(sizeof(int) * n);
  q->avail = q->step = n;
  q->size = 1;
  return q;
}


/********************************************************************
 * Destructor for queue structures
 *******************************************************************/
void pqdestroy(struct pqueue *q)
{
  free(q->cells);
  free(q->priorities);
  free(q);
}


/********************************************************************
 * pqinsert: insert an item into the queue.
 *
 * q is the pointer to a priority queue.
 * d is the datum to be inserted.
 ********************************************************************/
void pqinsert(struct pqueue *q, PQDATUM datum, int datum_priority)
{
  int i, newsize;

  assert(q);

  /* allocate more memory if necessary */
  if (q->size >= q->avail) {
    PQDATUM *tmp_d;
    int *tmp_p;
    newsize = q->size + q->step;
    freelog(LOG_NORMAL, "Resizing queue to %d", newsize);
    tmp_d = fc_realloc(q->cells, sizeof(PQDATUM) * newsize);
    tmp_p = fc_realloc(q->priorities, sizeof(int) * newsize);

    q->cells = tmp_d;
    q->priorities = tmp_p;
    q->avail = newsize;
  }

  /* insert item */
  i = q->size++;
  while (i > 1 && q->priorities[i / 2] < datum_priority) {
    q->cells[i] = q->cells[i / 2];
    q->priorities[i] = q->priorities[i / 2];
    i /= 2;
  }
  q->cells[i] = datum;
  q->priorities[i] = datum_priority;
}


/*******************************************************************
 * pqremove: remove the highest-ranking item from the queue.
 *
 * p is the pointer to a priority queue.
 *
 * Return value:
 *    non-NULL    The value of the item that has been removed.
 *    NULL        No item could be removed. Either the queue pointer
 *                provided was NULL, or the queue was empty.
 *******************************************************************/
PQDATUM pqremove(struct pqueue * q)
{
  PQDATUM tmp;
  int tmp_priority;
  PQDATUM top;
  int i = 1, j;

  assert(q->size <= q->avail);
  if (!q || q->size == 1)
    return PQDATUM_NULL;
  top = q->cells[1];
  q->size--;
  tmp = q->cells[q->size];
  tmp_priority = q->priorities[q->size];
  while (i <= q->size / 2) {
    j = 2 * i;
    if (j < q->size && q->priorities[j] < q->priorities[j + 1]) {
      j++;
    }
    if (q->priorities[j] <= tmp_priority) {
      break;
    }
    q->cells[i] = q->cells[j];
    q->priorities[i] = q->priorities[j];
    i = j;
  }
  q->cells[i] = tmp;
  q->priorities[i] = tmp_priority;
  return top;
}



/* --- The conversion methods (x,y) <--> xy --- */

/*******************************************************************
 * (x,y) --> xy
 *******************************************************************/
#define  get_xy_from_x_and_y(x, y) (y * MAX_XSIZE + x)

/*******************************************************************
 * xy --> x
 *******************************************************************/
#define  get_x_from_xy(xy) (xy % MAX_XSIZE)

/*******************************************************************
 * xy --> y
 *******************************************************************/
#define get_y_from_xy(xy) (xy / MAX_XSIZE)



/*-------------------- Possible cost functions -------------------*/
/* SINGLE_MOVE cost function */
int single_move(pf_map_t the_map, int x, int y, int dir)
{
  return SINGLE_MOVE;
}

/* SINGLE_MOVE cost function for SEA_MOVING */
int single_seamove(pf_map_t the_map, int x, int y, int dir)
{
  /* -3 means ships can move between */
  if ( map_get_tile(the_map->x, the_map->y)->move_cost[dir] == -3)
    return SINGLE_MOVE;
  else
    return PF_MAXCOST;
}

/* LAND_MOVE cost function for a unit */
int normal_move_unit(pf_map_t the_map, int x1, int y1, int dir)
{
  struct tile *ptile = map_get_tile(the_map->x, the_map->y);
  int t1 = map_get_terrain(x1, y1);
  int move_cost = PF_MAXCOST;

  if (t1 == T_OCEAN) {
    if (ground_unit_transporter_capacity(x1, y1, 
                                         the_map->params->owner) > 0)
      move_cost = SINGLE_MOVE;
    else 
      move_cost = PF_MAXCOST;
  } else if (ptile->terrain == T_OCEAN) {
    move_cost = get_tile_type(t1)->movement_cost * SINGLE_MOVE;
  } else {
    move_cost = ptile->move_cost[dir];
  }
  return move_cost;
}

/* NOT TESTED */
/* IGTER_MOVE cost function for a unit */
int igter_move_unit(pf_map_t the_map, int x1, int y1, int dir)
{
  struct tile *ptile = map_get_tile(the_map->x, the_map->y);
  int move_cost;

  if (map_get_terrain(x1, y1) == T_OCEAN) {
    if (ground_unit_transporter_capacity(x1, y1, 
                                         the_map->params->owner) > 0)
      move_cost = SINGLE_MOVE;
    else
      move_cost = PF_MAXCOST;
  } else if (ptile->terrain == T_OCEAN) {
    move_cost = MOVE_COST_ROAD;
  } else {
    /* Why is it SINGLE_MOVE below? */
    /* Or why is it MOVE_COST_ROAD above? */
    move_cost = (ptile->move_cost[dir] ? SINGLE_MOVE : 0);
  }
  return move_cost;
}

/* NOT TESTED */
/* LAND_MOVE cost function for a city ?? */
int normal_move_city(pf_map_t the_map, int x1, int y1, int dir)
{
  /* have no idea what code below is for */

  if (map_get_terrain(x1, y1) == T_OCEAN)
    return PF_MAXCOST;
  else {
    struct tile *ptile = map_get_tile(the_map->x, the_map->y);
    int tmp = map_get_tile(x1, y1)->move_cost[DIR_REVERSE(dir)];
    return (ptile->move_cost[dir] + tmp +
	    (ptile->move_cost[dir] > tmp ? 1 : 0))/2;
  }
}

/* --------------------- extra cost functions --------------------- */

/*********************************************************************
 * An example of extra cost fn: will avoid plains if possible 
 *********************************************************************/
int avoid_plains(int x, int y, enum known_type known, void *data)
{
  enum tile_terrain_type ter = map_get_terrain(x, y);

  switch (ter){
  case T_PLAINS:
    return 3;
  default:
    return 0;
  }
}

/* -------------------- tile status functions --------------------- */
/*********************************************************************
 * Will not go into mountains and will not pass through hills
 *********************************************************************/
enum tile_behaviour avoid_hills_TB(int x, int y, 
                                   enum known_type known, void *data) 
{
  switch (map_get_terrain(x, y)){
  case T_HILLS:
    return TB_NOTHROUGH;
  case T_MOUNTAINS:
    return TB_NOENTRY;
  default:
    return TB_OK;
  }
}

/*********************************************************************
 * Strict trireme function: will never go into the unknown and 
 * away from shore.
 *********************************************************************/
enum tile_behaviour strict_trireme(int x, int y, 
                                   enum known_type known, void *data)
{
  if (known != TILE_UNKNOWN && is_coastline(x,y)) {
    return TB_OK;
  }
  return TB_NOENTRY;
}

/* -------------------- manipulating the cost --------------------- */
/********************************************************************
 * Number of turns requite to reach node
 ********************************************************************/
#define get_turn(map, node) \
  ((node).cost / (map)->params->move_rate)

/********************************************************************
 * Moves left after node is reached
 ********************************************************************/
#define get_moves_left(map, node) \
  ((map)->params->move_rate \
   - ((node).cost % (map)->params->move_rate))

/********************************************************************
 * Adjust BMC to reflect the turn mode and the move_rate.
 * The result is AMC.
 ********************************************************************/
#define adjust_cost(map, cost) \
  if (cost < PF_MAXCOST) { \
    switch (map->params->turn_mode) { \
    case TM_NONE: \
      break; \
    case TM_AVERAGE: \
      cost = MIN(cost, map->params->move_rate); \
      break; \
    case TM_MAXIMUM: \
      { \
        int moves_left = get_moves_left(map, map->lattice[map->xy]); \
        \
        cost = (cost > moves_left ? moves_left + cost : cost); \
        break; \
      } \
    case TM_MINIMUM: \
      { \
        int moves_left = get_moves_left(map, map->lattice[map->xy]); \
        \
        cost = (cost > moves_left ? moves_left : cost); \
        break; \
      } \
    } \
  }


/* -------------------- get_known function(s) ------------ */
/********************************************************************
 * Server-side get_known function 
 ********************************************************************/
enum known_type server_get_known(int x, int y, struct player *pplayer)
{
  if (!map_get_known(x, y, pplayer)) {
    return TILE_UNKNOWN;
  } else if (map_get_known_and_seen(x, y, pplayer)) { 
    return TILE_KNOWN;
  } else {
    return TILE_KNOWN_FOGGED;
  }
}

/******************************************************************
 * Calculates cached values of the target node: 
 * node_known_type and zoc
 ******************************************************************/
static void pf_init_node(pf_map_t the_map, pf_node_t the_node, 
                         int x, int y)
{
  struct pf_parameter *params = the_map->params;

  the_node->status = NS_NEW;

  the_node->cost = PF_MAXCOST;

  /* Establish the "known" status of the_node */
  if (params->get_known) {
    the_node->node_known_type 
      = (*(params->get_known))(x, y, params->owner);
  } else {
    the_node->node_known_type = TILE_KNOWN;
  }

  /* Establish the tile behaviour */
  if (params->get_TB) {
    the_node->behaviour
      = (*(params->get_TB))(x, y, the_node->node_known_type, 
                            params->get_TB_data);
  } else {
    /* The default */
    the_node->behaviour = TB_OK;
  }

  if (the_map->params->zoc_used) {
    struct tile *the_tile = map_get_tile(x, y);
    
    the_node->is_my_zoc = (the_tile->city
                           || the_tile->terrain == T_OCEAN
		       || is_my_zoc(params->owner, x, y));
    the_node->is_allied_tile =
	(is_allied_unit_tile(the_tile, params->owner) != NULL);
  }

  /* Evaluate the extra cost of the destination */
  if (params->get_ECOT) {
    the_node->extra 
      = (*(params->get_ECOT))(x, y, the_node->node_known_type, 
                              params->get_ECOT_data);
  } else {
    the_node->extra = 0;
  }

}


/* -------------------- generating the map ------------------------ */

/*****************************************************************
 * Primary method for iterative path-finding.
 * Plan: 1. Process previous position
 *       2. Get new nearest position and return it
 *****************************************************************/
bool pf_iterate_map(pf_map_t the_map)
{
  xy_t xy;
  pf_node_t node_xy = &the_map->lattice[the_map->xy];

  node_xy->status = NS_PROCESSED;

  /* There is no exit from NOTHROUGH tiles! */
  if (node_xy->behaviour != TB_NOTHROUGH) {

    /* Processing Stage */
    /* The previous position is contained in {x,y} fields of the_map */
    
    adjc_dir_iterate(the_map->x, the_map->y, x1, y1, dir) {
      xy_t xy1 = get_xy_from_x_and_y(x1, y1);
      pf_node_t node_xy1 = &the_map->lattice[xy1];
      int cost;
      int extra = 0;
      
      if (node_xy1->status == NS_UNINIT) {
        pf_init_node(the_map, node_xy1, x1, y1);
      }
      
      /* Can we enter this tile at all? */
      if(node_xy1->behaviour == TB_NOENTRY) {
        continue;
      }
      
      /* Only update if this might shorten the paths */
      if (node_xy1->cost <= node_xy->cost) {
        continue;
      }
      
      /* Evaluate the cost of the move */
      cost = (*(the_map->get_cost))(the_map, x1, y1, dir);
      adjust_cost(the_map, cost);
      
      assert( cost > 0 ); /* No special cases yet */
      
      /* Total cost at xy1 */
      cost += node_xy->cost;
      /* Is the cost reasonable? */
      if (cost >= PF_MAXCOST) {
        continue;
      } 
      
      /* Evaluate the extra cost of the destination,
       * if it's relevant */
      if (the_map->params->get_ECOT) {
        extra = node_xy1->extra;
        extra += node_xy->extra_cost;
      }
      
      /* Update costs and add to queue, if we found a better route to xy1. */
      if (cost < node_xy1->cost 
          || (cost == node_xy1->cost && extra < node_xy1->extra_cost)) {
        int cost_of_path 
          = (the_map->params->get_COP 
             ? (*(the_map->params->get_COP))(cost, extra, 
                                           the_map->params->get_COP_data)
             : PF_COST_FACTOR * cost + extra );
        node_xy1->extra_cost = extra;
        node_xy1->cost = cost;
        node_xy1->dir_to_here = dir;
        pqinsert(the_map->queue, xy1, -cost_of_path);
      }
      
    } adjc_dir_iterate_end;
  }

  /* Get the next nearest node */
  do {
    xy = pqremove(the_map->queue);
    /* Discard if this node has already been processed */
  } while (xy != PQDATUM_NULL && the_map->lattice[xy].status != NS_NEW);

  if (xy == PQDATUM_NULL) {
    return FALSE;
  }
  the_map->xy = xy;
  the_map->x = get_x_from_xy(xy);
  the_map->y = get_y_from_xy(xy);

  return TRUE;
}

/******************************************************************
 * Allocates the memory for the map.  No initialization.
 ******************************************************************/
pf_map_t pf_create_map()
{
  pf_map_t the_map = fc_malloc(sizeof(struct pf_map));

  the_map->lattice = 
    fc_calloc(MAX_XSIZE * MAX_YSIZE, sizeof(struct pf_node));
  the_map->queue = pqcreate(INIT_ITEMS);

  return the_map;
}

/*****************************************************************
 * Create and fill the default parameter structure for a unit
 *****************************************************************/
struct pf_parameter *pf_create_unit_params(struct unit *punit)
{
  struct pf_parameter *params = fc_calloc(1, sizeof(struct pf_parameter));

  params->move_rate = unit_type(punit)->move_rate;
  params->start_x = punit->x;
  params->start_y = punit->y;
  params->owner = unit_owner(punit);
  params->moves_left_initially = punit->moves_left;

  return params;
}

/*****************************************************************
 * Creates the map and does minimal initializations.  
 * For debugging period only.  
 * To mature into pf_get_map eventually.
 *****************************************************************/
pf_map_t pf_init_map(struct unit *punit)
{
  pf_map_t the_map = pf_create_map();

  the_map->params = pf_create_unit_params(punit);

  the_map->params->turn_mode = TM_MAXIMUM;
  the_map->params->get_known = server_get_known;
  the_map->params->zoc_used = FALSE;

  the_map->x = the_map->params->start_x;
  the_map->y = the_map->params->start_y;
  the_map->xy = get_xy_from_x_and_y(the_map->x, the_map->y);

  if (is_ground_unit(punit)) {
    the_map->get_cost = normal_move_unit;
    the_map->params->get_TB = avoid_hills_TB;
    the_map->params->get_ECOT = avoid_plains;
  } else if (is_sailing_unit(punit)) {
    the_map->get_cost = single_seamove;
    if (unit_flag(punit, F_TRIREME)) {
      the_map->params->get_TB = strict_trireme;
    }
  } else {
    /* Yet unsupported */
    assert(0);
  }

  pf_init_node(the_map, &the_map->lattice[the_map->xy], 
               the_map->x, the_map->y);

  the_map->lattice[the_map->xy].cost = the_map->params->move_rate 
    - the_map->params->moves_left_initially;
  the_map->lattice[the_map->xy].extra_cost = 0;

  return the_map;
}

/*********************************************************************
 * After usage the map must be destroyed.
 *********************************************************************/
void pf_destroy_map(pf_map_t the_map)
{
  free(the_map->lattice);
  free(the_map->params);   /* Remove it later */
  pqdestroy(the_map->queue);
  free(the_map);
}


/*---------------------- Lifting info from the map ----------------*/
/*******************************************************************
 * x of the current node
 *******************************************************************/
int pf_get_current_x(const pf_map_t the_map)
{
  return the_map->x;
}

/*******************************************************************
 * y of the current node
 *******************************************************************/
int pf_get_current_y(const pf_map_t the_map)
{
  return the_map->y;
}

/*******************************************************************
 * cost of the shortest path to the current node
 *******************************************************************/
int pf_get_current_cost(const pf_map_t the_map)
{
  return the_map->lattice[the_map->xy].cost;
}

/*******************************************************************
 * extra cost of the shortest path to the current node
 *******************************************************************/
int pf_get_current_extra_cost(const pf_map_t the_map)
{
  return the_map->lattice[the_map->xy].extra_cost;
}

/*******************************************************************
 * Read off the path to the current node
 *******************************************************************/
bool pf_get_current_path(const pf_map_t the_map, struct pf_path *path)
{
  struct pf_position path_rev[MAX_PATH_LENGTH + 1];
  int i;
  int x = the_map->x, y = the_map->y;
  enum direction8 dir_next = -1;

  for (i = 0; i <= MAX_PATH_LENGTH; i++) {
    pf_node_t node_xy = &the_map->lattice[get_xy_from_x_and_y(x, y)];
    path_rev[i].x = x;
    path_rev[i].y = y;
    path_rev[i].cost = node_xy->cost;
    path_rev[i].TEC = node_xy->extra_cost;
    if (the_map->params->turn_mode == TM_MAXIMUM ||
        the_map->params->turn_mode == TM_MINIMUM) {
      path_rev[i].turn = get_turn(the_map, *node_xy);
      path_rev[i].moves_left = get_moves_left(the_map, *node_xy);
    } else {
      path_rev[i].turn = -1;
      path_rev[i].moves_left = -1;
    }

    path_rev[i].direction_for_next_position = dir_next;
    dir_next = node_xy->dir_to_here;

    if (same_pos(x, y, 
                 the_map->params->start_x, the_map->params->start_y)) {
      /* Ah-ha, reached the starting point! */
      int j;

      path->positions_used = i + 1;
      path->is_valid = TRUE;
      
      /* Actually filling the path */
      for(j = 0; i>=0; i--, j++) {
        memcpy(&path->positions[j], &path_rev[i], 
               sizeof(struct pf_position));
      }
      
      return TRUE;
    }

    /* Need to step further back */
    if (!STEP(x, y, DIR_REVERSE(dir_next))) {
      freelog(LOG_NORMAL, "pf_get_current_path: wrong directions recorded!");
      path->is_valid = FALSE;
      return FALSE;
    }
  }

  freelog(LOG_NORMAL, "pf_get_current_path: path too long");
  path->is_valid = FALSE;
  return FALSE;
}

/************************************************************************
 * Printing a path
 ************************************************************************/
void pf_print_path(const struct pf_path *path)
{
  int i;

  freelog(LOG_NORMAL, "PF: path (at %p) is %s", path,
	  path->is_valid ? "valid" : "invalid");
  if (!path->is_valid) {
    return;
  }
  freelog(LOG_NORMAL, "PF:  consists of %d positions:",
	  path->positions_used);
  
  for (i = 0; i < path->positions_used; i++) {
    freelog(LOG_NORMAL,
	    "PF:   %2d/%2d: (%2d,%2d) cost=%2d (%2d, %d) EC=%d",
	    i + 1, path->positions_used, 
            path->positions[i].x, path->positions[i].y,
	    path->positions[i].cost, path->positions[i].turn, 
            path->positions[i].moves_left, path->positions[i].TEC);
  }
}


/*********************************************************************
 * Time the warmap execution.
 *********************************************************************/
void time_path_finding(struct player *pplayer)
{
  /*  struct tms *start, *finish;  */
  clock_t start, finish;
  int i;

  start = clock();

  for(i = 1; i < 1000; i++) {
    unit_list_iterate(pplayer->units, punit) {
      pf_map_t test_map = pf_init_map(punit);
      /* freelog(LOG_NORMAL, "%s (%d,%d)", 
         unit_type(punit)->name, punit->x, punit->y);  */
      while (pf_iterate_map(test_map)) {
        /*        freelog(LOG_NORMAL, "Processing (%d,%d)",
                  pf_get_current_x(test_map), pf_get_current_y(test_map)); */
        /* empty */
      }
      pf_destroy_map(test_map);
    } unit_list_iterate_end;
  }

  finish = clock();

  freelog(LOG_NORMAL, "Secs used: %f", 
          (float) (finish - start) / CLOCKS_PER_SEC );

}
