/********************************************************************** 
 Freeciv - Copyright (C) 2002 - The Freeciv Project
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/
#ifndef FC__PATH_FINDING_H
#define FC__PATH_FINDING_H

#include "map.h"
#include "unit.h"
#include "unittype.h"
#include "terrain.h"

/*
 * Functions in this file help to find a path from A to B.
 *
 * Terms used:
 *  - step: one movement step which brings us from one tile to an adjacent one
 *  - turn: obvious
 *  - path: a list of steps which leads from the start to the end
 *  - basic movement cost (BMC) (of a step): the basic movement costs
 *  of the step as returned by map_move_cost.  BMC does not depend on the 
 *  condition of the unit.
 *  - adjusted movement cost (AMC): BMC adjusted to take into account the
 *  move_rate of the unit and the turn_mode (see below).
 *  - extra cost of a step (ECS): sum of various extra costs at the target
 *  tile (all cost have to be >=0)
 *  - cost of a path (COP): 
 *        COP = (sum of AMCs) * PF_AMC_FACTOR + (sum of ECs),
 *  where the sum is over all step of the path.
 *  - best path: a path which has the minimal COP. If two paths have
 *  equal COP the behavior is undefined.
 */

/*
 * Opaque type(s).
 */
typedef struct pf_map *pf_map_t;

#define PF_MAX_COST    63

/*
 * Default scaling factor for the BMC wrt extra cost.
 */
#define PF_COST_FACTOR	1000

/* MAXCOST is (uchar_t)(-1) - some initializations depend on this */
#define PF_MAXCOST 255

/* 
 * Generic extra cost function:
 * (x,y) is the target tile.
 */
typedef int (EXTRAFN)(pf_map_t the_map, int x, int y);

/* 
 * The return value for get_tile_status callback
 * NOENTRY 
 * (see http://www.roads.dft.gov.uk/roadsafety/hc/sign043.shtml)
 * means the tile will be ignored by the path-finding engine
 * NOTHROUGH 
 * (see http://www.roads.dft.gov.uk/roadsafety/hc/sign164.shtml)
 * means no paths through the tile will be considered, but a path to 
 * the tile will be found, if possible
 * OK means no restrictions
 * (see http://www.roads.dft.gov.uk/roadsafety/hc/sign003.shtml)
 */
enum tile_behaviour {
  TB_NOENTRY,
  TB_NOTHROUGH, 
  TB_OK
};

/*
 * If unit has less than full movepoints and tries to undertake a move 
 * which costs more than the unit has, the move succeeds with probability
 * (current_move_points / BMC).  This probabilistic nature of the
 * movement means that we are generally unable to predict the number of 
 * turns needed to reach the destination.  We can, however, find the 
 * minimum, maximum and some sort of average time to the destination.
 * In MINIMUM mode we assume that if current_move_points > 0, 
 * the move will always succeed.
 * In MAXIMUM mode we assume that if current_move_points < full_move_points
 * and current_move_points < BMC, the move will always fail.
 * In AVERAGE mode we just add up all the BMC and divide by 
 * full_move_points.
 * Note: In CivIII the movement is no longer random, it is done according to
 * the MINIMUM mode.
 */
enum turn_mode {
  TM_NONE, TM_MINIMUM, TM_AVERAGE, TM_MAXIMUM
};

/* 
 * Creates and initializes the map 
 */
pf_map_t pf_init_map(struct unit *punit);

/* 
 * Performs one iteration on the map.  The current posistion and 
 * the costs can be read using following utility functions 
 */
bool pf_iterate_map(pf_map_t pf_map);

/*
 * After usage the map should be destroyed.
 */
void pf_destroy_map(pf_map_t pf_map);


/*
 * Provide info about current position of the map iteration.
 */
int pf_get_current_x(const pf_map_t);
int pf_get_current_y(const pf_map_t);
int pf_get_current_cost(const pf_map_t);
int pf_get_current_extra_cost(const pf_map_t);

/*
 * Maximal number of steps in a path.
 */
#define MAX_PATH_LENGTH	100

struct pf_path {
  /*
   * Is this path valid? Is set to FALSE if there was no path found.
   */
  bool is_valid;

  /*
   * Number of positions used in the array "positions".
   */
  int positions_used;

  struct pf_position {
    int x, y;
    int turn, moves_left;
    int total_BMC;
    /* TEC required to reach this position */
    int TEC;

    /* Total cost required to reach this position */
    int cost;
    enum direction8 direction_for_next_position;
  } positions[MAX_PATH_LENGTH + 1];
};


/*
 * Construct the whole path to the current position of the map iteration 
 * and write it into the given path.
 */
bool pf_get_current_path(pf_map_t pf_map, struct pf_path *path);

void pf_print_path(const struct pf_path *path);

/* 
 * Print the system time used by a big map calculation
 */
void time_path_finding(struct player *pplayer);


struct pf_parameter {
  /*
   * Definition of the state of a virtual unit unit at the start.
   */
  int start_x, start_y;
  int moves_left_initially;
  struct player *owner;

  enum unit_move_type move_type;
  /* only F_IGZOC and F_IGTER are used currently */
  bv_flags flags;
  int move_rate;

  /* 
   * Only affects the BMC. If set the FALSE the BMC is calculated to
   * simulate a path from (start_x, start_y) to the location (which is
   * returned by pf_get_next_position or passed to pf_get_path). If
   * set to TRUE the BMC will be calculated to simulate a path from
   * the location to the starting map position (start_x,start_y).
   */
  bool move_backward;

  bool zoc_used;

  /*
   * Determines how turns and moves_left are calculated.
   */
  enum turn_mode turn_mode;

  /*
   * Callback to query the known state of a tile for the given
   * player. Should take into account the handicaps. If NULL
   * TILE_KNOWN is assumed.
   */
  enum known_type (*get_known) (int, int, struct player *);

  /*
   * GOTO_MOVE_STRAIGHTEST will set the BMC to constant 1. Even for
   * unknown tiles.
   */
  enum goto_move_restriction restriction;

  /*
   * Passed as last parameter to get_TB.
   */
  void *get_TB_data;

  /*
   * Callback which determines the behavior of a tile. If NULL
   * TB_NORMAL is assumed. Parameters are x, y, known and
   * get_TB_data. It can be assumed that the implementation of
   * path_finding.h will cache this value.
   */
  enum tile_behaviour (*get_TB) (int, int, enum known_type, void *);

  /*
   * Passed as last parameter to get_ECOT.
   */
  void *get_ECOT_data;

  /*
   * Callback which can be used to provide extra costs depending on
   * the tile. Can be NULL. Parameters are x, y, known and
   * get_ECOT_data. It can be assumed that the implementation of
   * path_finding.h will cache this value.
   */
  int (*get_ECOT) (int, int, enum known_type, void *);

  /*
   * Passed as last parameter to is_position_safe.
   */
  void *is_position_dangerous_data;

  /*
   * If this callback is non-NULL and returns TRUE this position is
   * dangerous. The unit will never end a turn at a dangerous
   * position. Can be NULL. Parameter are x, y and
   * is_position_dangerous_data.
   */
   bool(*is_position_dangerous) (int, int, void *);

  /*
   * Passed as last parameter to get_cop.
   */
  void *get_COP_data;

  /*
   * If get_cop is NULL the default formula for COP is used, which is
   * COP = TEC + cost * PF_COST_FACTOR, where
   *    cost = BMC          (TM_NONE)
   *    cost = (turn + 1) * move_rate - moves_left (other cases).
   * 
   * If get_cop is non-NULL the get_cop function will be called and
   * returns the COP. Parameter are cost (see above),
   * TEC and get_cop_data.
   */
  int (*get_COP) (int, int, void *);
};


/* -------------------------------------------------------------- */
/* --------- From here down nothing is used / implemented ------- */
/* -------------------------------------------------------------- */

/*
 * Called only once per executable start.
 */
void pf_init(void);

/*
 * Returns a map which can be used to query for paths or to iterate
 * over all paths. Does not perform any computations itself, just sets
 * everything up.
 */
pf_map_t pf_get_map(const struct pf_parameter *const parameter);

/*
 * Tries to find the best path in the given map to the given
 * destination position. The path will be written in the given struct
 * pf_path. The caller should test path->found_a_valid.
 */
void pf_get_path(pf_map_t pf_map, int end_x, int end_y, struct pf_path *path);

/*
 * Returns the last position of the given path.
 */
struct pf_position *last_position_on_path(struct pf_path *path);


#endif

